'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

class Context {
    constructor() {
        this.readyQueue = [];
        this.isSafari = typeof AudioContext === 'undefined' && !!webkitAudioContext;
        this.initialized = false;
    }
    start() {
        if (!this.started) {
            this.started = new Promise((resolve, reject) => {
                try {
                    this.context = new (this.isSafari ? webkitAudioContext : AudioContext)();
                    this.context.onstatechange = () => {
                        if (this.context.state === 'running') {
                            this.readyQueue.forEach(f => f(this.context));
                            this.readyQueue = [];
                            resolve(this.context);
                            this.initialized = true;
                        }
                    };
                    this.context.resume();
                }
                catch (_a) {
                    reject();
                }
            });
        }
        return this.started;
    }
    ready(func) {
        if (this.context) {
            func(this.context);
        }
        else {
            this.readyQueue.push(func);
        }
    }
    getTime() {
        if (this.context) {
            return this.context.currentTime;
        }
        return 0;
    }
}

const defaultContext = new Context();

class Node {
    constructor() {
        this.parents = [];
        this.children = [];
        this.autoInvalidateChildren = true;
    }
    connect(...children) {
        children.forEach(child => {
            if (!this.children.includes(child)) {
                child.addParent(this);
                this.children.push(child);
            }
        });
    }
    disconnect(...children) {
        children.forEach(child => {
            if (this.children.includes(child)) {
                child.removeParent(this);
                this.children.splice(this.children.findIndex(c => c === child), 1);
            }
        });
    }
    addParent(parent) {
        if (!this.parents.includes(parent)) {
            this.parents.push(parent);
            clearTimeout(this.invalidationTimeout);
            this.invalidationTimeout = setTimeout(() => this.invalidateConnections(), 0);
        }
    }
    removeParent(parent) {
        if (this.parents.includes(parent)) {
            this.parents.splice(this.parents.findIndex(p => p === parent), 1);
            clearTimeout(this.invalidationTimeout);
            this.invalidationTimeout = setTimeout(() => this.invalidateConnections(), 0);
        }
    }
    invalidateConnections() {
        if (this.autoInvalidateChildren && this.children.length > 0) {
            this.children.forEach(c => c.invalidateConnections());
        }
    }
    getOutputAudioNodes() {
        return this.parents.map(c => c.getOutputAudioNodes()).flat();
    }
    destroy() {
        this.disconnect(...this.children);
        this.parents.forEach(p => p.disconnect(this));
    }
}

class OperativeNode extends Node {
    constructor({ context = defaultContext, } = {}) {
        super();
        this.upstreamNodes = [];
        this.autoInvalidateChildren = false;
        this.context = context;
    }
    rebuildUpstreamConnections() {
        const currentNodes = this.parents.map(p => p.getOutputAudioNodes()).flat();
        const disconnectedNodes = this.upstreamNodes.filter(n => !currentNodes.includes(n));
        const newNodes = currentNodes.filter(n => !this.upstreamNodes.includes(n));
        disconnectedNodes.forEach(n => {
            this.upstreamNodes.splice(this.upstreamNodes.findIndex(n2 => n2 === n), 1);
            n.disconnect(this.node);
        });
        newNodes.forEach(n => {
            this.upstreamNodes.push(n);
            n.connect(this.node);
        });
    }
    invalidateConnections() {
        this.rebuildUpstreamConnections();
        super.invalidateConnections();
    }
    getOutputAudioNodes() {
        return this.node ? [this.node] : [];
    }
}

class Destination extends OperativeNode {
    constructor(props) {
        super(props);
        this.context.ready(ctx => {
            this.node = ctx.destination;
        });
    }
    connect() {
        throw new Error('A Destination can\'t have children.');
    }
    disconnect() {
        throw new Error('A Destination can\'t have children.');
    }
}

const defaultDestination = new Destination({
    context: defaultContext,
});

class Bus extends Node {
    constructor({ effects = [], }) {
        super();
        this.effects = effects;
        for (let i = 0; i < effects.length - 1; i += 1) {
            effects[i].connect(effects[i + 1]);
        }
    }
    addParent(parent) {
        super.addParent(parent);
        if (this.effects.length > 0) {
            parent.connect(this.effects[0]);
        }
    }
    removeParent(parent) {
        if (this.effects.length > 0) {
            parent.disconnect(this.effects[0]);
        }
        super.removeParent(parent);
    }
    getOutputAudioNodes() {
        if (this.effects.length > 0) {
            return this.effects[this.effects.length - 1].getOutputAudioNodes();
        }
        return super.getOutputAudioNodes();
    }
}

class Timeline {
    constructor({ bpm = 60, mesure = 4 / 4, lookahead = .25, effects = [], context = defaultContext, destination = defaultDestination, }) {
        this.started = false;
        this.startTime = 0;
        this.nextPartStart = 0;
        this.queue = [];
        this.context = context;
        this.barDuration = 60 / (bpm / (mesure * 4));
        this.lookahead = lookahead;
        this.bus = new Bus({
            effects,
        });
        this.bus.connect(destination);
    }
    play(sound) {
        sound.bindTimeline(this);
        sound.play();
    }
    cue(part) {
        this.queue.push(part);
        part.bindTimeline(this);
        if (!this.started) {
            this.started = true;
            this.startTime = this.context.context.currentTime
                + this.lookahead * this.barDuration;
            this.nextPartStart = this.startTime;
            this.updateState();
        }
    }
    updateState() {
        if (this.queue.length > 0) {
            this.currentPart = this.queue.shift();
        }
        else if (!this.currentPart || !this.currentPart.loop) {
            this.currentPart.bus.disconnect(this.bus);
            this.currentPart = null;
        }
        if (this.currentPart) {
            this.currentPart.play(this.nextPartStart);
            this.nextPartStart += this.currentPart.duration * this.barDuration;
            setTimeout(() => this.updateState(), this.getTimeoutForTime(this.nextPartStart));
        }
        else {
            this.started = false;
        }
    }
    getTimeoutForTime(time) {
        return Math.max(0, ((time - this.context.getTime()) - this.barDuration * this.lookahead) * 1000);
    }
}

/*! *****************************************************************************
Copyright (c) Microsoft Corporation.

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
***************************************************************************** */

function __rest(s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
}

function loadAudio(input, context) {
    console.log(context);
    return new Promise((resolve, reject) => {
        if (input instanceof AudioBuffer) {
            resolve(input);
        }
        else if (input instanceof ArrayBuffer) {
            context.ready(() => {
                context.context.decodeAudioData(input, (audio) => {
                    resolve(audio);
                });
            });
        }
        else {
            fetch(input).then(response => response.arrayBuffer()).then((buffer) => {
                context.ready((ctx) => {
                    ctx.decodeAudioData(buffer, audio => resolve(audio), () => {
                        console.error(`Error decoding file ${input}`);
                        reject();
                    });
                });
            }).catch(() => {
                console.error(`Error loading file ${input}`);
                reject();
            });
        }
    });
}

class Source extends OperativeNode {
    constructor(_a = {}) {
        var { src, gain = 1, onAutoDestroy, autoDestroy = true } = _a, rest = __rest(_a, ["src", "gain", "onAutoDestroy", "autoDestroy"]);
        super(rest);
        this.loaded = false;
        this.hasExternalBuffer = false;
        this.src = src;
        this.gain = gain;
        this.context.ready((ctx) => {
            this.node = ctx.createGain();
            this.node.gain.value = 0;
            this.sourceNode = ctx.createBufferSource();
            if (autoDestroy) {
                this.sourceNode.onended = () => {
                    this.destroy();
                    if (onAutoDestroy)
                        onAutoDestroy();
                };
            }
            this.sourceNode.connect(this.node);
        });
    }
    load() {
        if (!this.loaded) {
            this.loaded = true;
            return loadAudio(this.src, this.context).then((audio) => {
                this.context.ready(() => {
                    this.sourceNode.buffer = audio;
                });
            });
        }
        return Promise.resolve();
    }
    setBuffer(audio) {
        this.sourceNode.buffer = audio;
        this.loaded = true;
        this.hasExternalBuffer = true;
    }
    play(time = 0) {
        this.sourceNode.start(time);
        this.node.gain.setTargetAtTime(this.gain, time, .02);
    }
    destroy() {
        this.sourceNode.stop();
        this.sourceNode.disconnect(this.node);
        if (!this.hasExternalBuffer) {
            this.loaded = false;
            this.sourceNode.buffer = null;
        }
        super.destroy();
    }
    addParent() {
        throw new Error('A Source cannot have parents.');
    }
    removeParent() {
        throw new Error('A Source cannot have parents.');
    }
}

class Container {
    constructor({ effects = [], }) {
        this.bus = new Bus({
            effects,
        });
    }
    bindTimeline(timeline) {
        this.timeline = timeline;
    }
    destroy() {
        this.bus.destroy();
    }
}

class Playable extends Container {
    constructor(_a) {
        var { src, gain, load = true } = _a, rest = __rest(_a, ["src", "gain", "load"]);
        super(rest);
        this.sources = [];
        this.src = src;
        this.gain = gain;
        this.autoLoad = load;
    }
    load() {
        if (!this.buffer) {
            return loadAudio(this.src, this.context || defaultContext).then(buffer => {
                this.buffer = buffer;
            });
        }
        return Promise.resolve();
    }
    createSource() {
        const source = new Source({
            context: this.context,
            gain: this.gain,
            onAutoDestroy: () => {
                source.disconnect(this.bus);
                this.sources.splice(this.sources.findIndex(s => s === source), 1);
            },
        });
        source.setBuffer(this.buffer);
        source.connect(this.bus);
        this.sources.push(source);
        return source;
    }
    bindTimeline(timeline) {
        super.bindTimeline(timeline);
        this.context = timeline.context;
        if (this.autoLoad)
            this.load();
    }
    destroy() {
        this.sources.forEach(s => s.destroy());
        super.destroy();
    }
}

class Track extends Playable {
    constructor(_a) {
        var { duration, loop = false } = _a, rest = __rest(_a, ["duration", "loop"]);
        super(rest);
        this.duration = duration;
        this.loop = loop;
    }
    play(start = 0) {
        if (this.buffer) {
            const time = this.timeline.context.getTime();
            const startTimes = this.loop
                ? Array(Math.round(this.part.duration / this.duration))
                    .fill(0)
                    .map((_, i) => start + (i * this.duration * this.timeline.barDuration))
                : [start];
            startTimes.forEach((startTime) => {
                if (startTime < time) {
                    console.warn('Cue has already passed. Skipping...');
                }
                else {
                    const source = this.createSource();
                    source.play(startTime);
                }
            });
        }
    }
    bindPart(part) {
        this.part = part;
        this.bus.connect(part.bus);
    }
}

class Part extends Container {
    constructor({ tracks, effects, duration, loop = true, }) {
        super({ effects });
        this.tracks = tracks;
        this.duration = duration;
        this.loop = loop;
        tracks.forEach(t => t.bindPart(this));
    }
    bindTimeline(timeline) {
        super.bindTimeline(timeline);
        this.bus.connect(timeline.bus);
        this.tracks.forEach(t => t.bindTimeline(timeline));
    }
    play(time) {
        this.tracks.forEach(t => t.play(time));
    }
    load() {
        return Promise.all(this.tracks.map(t => t.load()));
    }
    destroy() {
        this.timeline.bus.disconnect(this.bus);
        this.tracks.forEach(t => t.destroy());
        super.destroy();
    }
}

class ActionSound extends Playable {
    constructor(_a) {
        var { quantize = 0 } = _a, rest = __rest(_a, ["quantize"]);
        super(rest);
        this.quantize = quantize;
    }
    play() {
        if (this.buffer) {
            const source = this.createSource();
            if (this.quantize > 0) {
                const time = this.timeline.context.getTime();
                const b = Math.ceil(time / (this.timeline.barDuration * this.quantize));
                source.play(b * this.timeline.barDuration * this.quantize);
            }
            else {
                source.play(0);
            }
        }
    }
    bindTimeline(timeline) {
        super.bindTimeline(timeline);
        this.bus.connect(timeline.bus);
    }
}

class EffectNode extends OperativeNode {
    constructor(factory, props) {
        super(props);
        this.context.ready((ctx) => {
            this.node = factory(ctx);
        });
    }
    getControllerTarget() {
        return this.node;
    }
}

class BiquadFilter extends EffectNode {
    constructor(_a = {}) {
        var { type = 'lowpass', frequency = 350, detune = 0, Q = 1, gain = 1 } = _a, rest = __rest(_a, ["type", "frequency", "detune", "Q", "gain"]);
        super((ctx) => {
            const node = ctx.createBiquadFilter();
            node.type = type;
            node.frequency.value = frequency;
            node.detune.value = detune;
            node.Q.value = Q;
            node.gain.value = gain;
            return node;
        }, rest);
    }
}

class Convolver extends EffectNode {
    constructor(_a) {
        var { ir, normalize = true } = _a, rest = __rest(_a, ["ir", "normalize"]);
        super((ctx) => {
            const node = ctx.createConvolver();
            node.normalize = normalize;
            return node;
        }, rest);
        loadAudio(ir, this.context).then((audio) => {
            this.context.ready(() => {
                this.node.buffer = audio;
            });
        });
    }
}

class Delay extends EffectNode {
    constructor(_a = {}) {
        var { delayTime = 0 } = _a, rest = __rest(_a, ["delayTime"]);
        super((ctx) => {
            const node = ctx.createDelay();
            node.delayTime.value = delayTime;
            return node;
        }, rest);
    }
}

class DynamicsCompressor extends EffectNode {
    constructor(_a = {}) {
        var { threshold = -24, knee = 30, ratio = 12, attack = .003, release = .25 } = _a, rest = __rest(_a, ["threshold", "knee", "ratio", "attack", "release"]);
        super((ctx) => {
            const node = ctx.createDynamicsCompressor();
            node.threshold.value = threshold;
            node.knee.value = knee;
            node.ratio.value = ratio;
            node.attack.value = attack;
            node.release.value = release;
            return node;
        }, rest);
    }
}

class Gain extends EffectNode {
    constructor(_a = {}) {
        var { gain = 1 } = _a, rest = __rest(_a, ["gain"]);
        super((ctx) => {
            const node = ctx.createGain();
            node.gain.value = gain;
            return node;
        }, rest);
    }
}

const EffectController = class EffectController {
    constructor({ effects, smoothing = .001, context = defaultContext, }) {
        this.effects = effects;
        this.smoothing = smoothing;
        this.context = context;
    }
    set(key, value) {
        if (this.context.initialized) {
            this.effects.forEach((effect) => {
                const n = effect.getControllerTarget();
                const param = n[key];
                param.cancelScheduledValues(0);
                param.setTargetAtTime(value, 0, this.smoothing);
            });
        }
    }
    connect(...nodes) {
        nodes.forEach((node) => {
            if (!this.effects.includes(node)) {
                this.effects.push(node);
            }
        });
    }
    disconnect(...nodes) {
        nodes.forEach(node => {
            if (this.effects.includes(node)) {
                this.effects.splice(this.effects.findIndex(n => n === node), 1);
            }
        });
    }
};

exports.ActionSound = ActionSound;
exports.BiquadFilter = BiquadFilter;
exports.Bus = Bus;
exports.Context = Context;
exports.Convolver = Convolver;
exports.Delay = Delay;
exports.Destination = Destination;
exports.DynamicsCompressor = DynamicsCompressor;
exports.EffectController = EffectController;
exports.Gain = Gain;
exports.Part = Part;
exports.Source = Source;
exports.Timeline = Timeline;
exports.Track = Track;
exports.defaultContext = defaultContext;
exports.defaultDestination = defaultDestination;
